package l15_collections;

import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;

// Collections.synchronizedMap() and ConcurrentHashMap both provide thread-safe
// operations on collections of data.

// The Collections utility class provides polymorphic algorithms that operate on collections
// and return wrapped collections. Its synchronizedMap() method provides thread-safe functionality.

// As the name implies, synchronizedMap() returns a synchronized Map backed by the Map that we
// provide in the parameter. To provide thread-safety, synchronizedMap() allows all accesses
// to the backing Map via the returned Map.

// The ConcurrentHashMap class is thread-safe. Therefore, multiple threads can operate
// on a single object with no complications.

// In ConcurrentHashMap, read operations are non-blocking, whereas write operations take
// a lock on a particular segment or bucket. The default bucket or concurrency level is 16,
// which means 16 threads can write at any instant after taking a lock on a segment or bucket.

// Collections.synchronizedMap() requires each thread to acquire a lock on the entire object
// for both read/write operations. By comparison, the ConcurrentHashMap allows threads to
// acquire locks on separate segments of the collection, and make modifications at the same time.


public class SynchronizedMapExample {

	private static final AtomicInteger atomicInteger = new AtomicInteger();

	private static void modifyWhileIterating(Map<Integer, String> map) {

		map.put(1, "Hello");
		map.put(2, "HashMap");
		System.out.println(map.getClass().getSimpleName());
		System.out.println(map);
		try {
			for (var k : map.keySet()) map.remove(k);
//			map.forEach((k, v) -> map.remove(k));
		} catch (ConcurrentModificationException e) {
			e.printStackTrace();
		}
		System.out.println(STR."size \{map.size()}");
		System.out.println();
	}

	private static void modifyByThreads(Map<Integer, String> map) {
		try (ExecutorService e = Executors.newFixedThreadPool(8)) {
			for (int i = 0; i < 10000; i++) {
				e.submit(() -> map.put(atomicInteger.incrementAndGet(), "Hello"));
			}
		}
		System.out.println(STR."\{map.getClass().getSimpleName()} final size is \{map.size()}");
	}

	public static void main(String... args) {

		modifyByThreads(new HashMap<>());
		modifyByThreads(Collections.synchronizedMap(new HashMap<>()));
		modifyByThreads(new ConcurrentHashMap<>());
		System.out.println("----------------------");
		modifyWhileIterating(new HashMap<>());
		modifyWhileIterating(Collections.synchronizedMap(new HashMap<>()));
		modifyWhileIterating(new ConcurrentHashMap<>());
	}
}
